@@ -1,7 +1,7 @@
 """ Simple interface for CardiffNLP twitter models. """
 import logging
 import os
-from typing import List
+from typing import List, Dict
 
 import torch
 
@@ -53,13 +53,21 @@ MODEL_LIST = {
 class Classifier:
 
     def __init__(self,
-                 model_name: str,
-                 max_length: int,
+                 model_name: str = None,
+                 max_length: int = 128,
                  multi_label: bool = False,
-                 use_auth_token: bool = False):
-        logging.debug(f'loading {model_name}')
-        self.config, self.tokenizer, self.model = load_model(
-            model_name, task='sequence_classification', use_auth_token=use_auth_token)
+                 use_auth_token: bool = False,
+                 loaded_model_config_tokenizer: Dict = None):
+        if loaded_model_config_tokenizer is not None:
+            assert all(i in loaded_model_config_tokenizer.keys() for i in ['model', 'config', 'tokenizer'])
+            self.config = loaded_model_config_tokenizer['config']
+            self.tokenizer = loaded_model_config_tokenizer['tokenizer']
+            self.model = loaded_model_config_tokenizer['model']
+        else:
+            assert model_name is not None, "model_name is required"
+            logging.debug(f'loading {model_name}')
+            self.config, self.tokenizer, self.model = load_model(
+                model_name, task='sequence_classification', use_auth_token=use_auth_token)
         self.max_length = max_length
         self.multi_label = multi_label
         self.id_to_label = {str(v): k for k, v in self.config.label2id.items()}
